#!/var/folders/nz/j6p8yfhx1mv_0grj5xl4650h0000gp/T/abs_a39nh5lowr/croot/gdal-split_1691162687573/_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_pla/bin/python

import sys

from osgeo.gdal import deprecation_warn

# import osgeo_utils.gdal_polygonize as a convenience to use as a script
from osgeo_utils.gdal_polygonize import *  # noqa
from osgeo_utils.gdal_polygonize import main

deprecation_warn("gdal_polygonize")
sys.exit(main(sys.argv))
